@@ -13,13 +13,11 @@
  * GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+ * License along with this program; if not,  see <http://www.gnu.org/licenses>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
 #define _LARGEFILE64_SOURCE
-#define _GNU_SOURCE
 #include <dirent.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -37,6 +35,7 @@
 #include <errno.h>
 
 #include "trace-cmd-local.h"
+#include "kbuffer.h"
 #include "list.h"
 
 #define MISSING_EVENTS (1 << 31)
@@ -54,6 +53,9 @@ struct page {
 	void			*map;
 	int			ref_count;
 	long long		lost_events;
+#if DEBUG_RECORD
+	struct pevent_record	*records;
+#endif
 };
 
 struct cpu_data {
@@ -64,16 +66,16 @@ struct cpu_data {
 	unsigned long long	size;
 	unsigned long long	timestamp;
 	struct list_head	pages;
-	struct record		*next;
+	struct pevent_record	*next;
 	struct page		*page;
+	struct kbuffer		*kbuf;
 	int			cpu;
-	int			index;
-	int			page_size;
 };
 
 struct tracecmd_input {
 	struct pevent		*pevent;
 	struct plugin_list	*plugin_list;
+	unsigned long		flags;
 	int			fd;
 	int			long_size;
 	int			page_size;
@@ -81,15 +83,77 @@ struct tracecmd_input {
 	int			cpus;
 	int			ref;
 	struct cpu_data 	*cpu_data;
+	unsigned long long	ts_offset;
+	char *			cpustats;
+
+	struct tracecmd_ftrace	finfo;
 
 	/* file information */
 	size_t			header_files_start;
 	size_t			ftrace_files_start;
 	size_t			event_files_start;
+	size_t			total_file_size;
 };
 
 __thread struct tracecmd_input *tracecmd_curr_thread_handle;
 
+void tracecmd_set_flag(struct tracecmd_input *handle, int flag)
+{
+	handle->flags |= flag;
+}
+
+void tracecmd_clear_flag(struct tracecmd_input *handle, int flag)
+{
+	handle->flags &= ~flag;
+}
+
+#if DEBUG_RECORD
+static void remove_record(struct page *page, struct pevent_record *record)
+{
+	if (record->prev)
+		record->prev->next = record->next;
+	else
+		page->records = record->next;
+	if (record->next)
+		record->next->prev = record->prev;
+}
+static void add_record(struct page *page, struct pevent_record *record)
+{
+	if (page->records)
+		page->records->prev = record;
+	record->next = page->records;
+	record->prev = NULL;
+	page->records = record;
+}
+static const char *show_records(struct list_head *pages)
+{
+	static char buf[BUFSIZ + 1];
+	struct pevent_record *record;
+	struct page *page;
+	int len;
+
+	memset(buf, 0, sizeof(buf));
+	len = 0;
+	list_for_each_entry(page, pages, list) {
+		for (record = page->records; record; record = record->next) {
+			int n;
+			n = snprintf(buf+len, BUFSIZ - len, " 0x%lx", record->alloc_addr);
+			len += n;
+			if (len >= BUFSIZ)
+				break;
+		}
+	}
+	return buf;
+}
+#else
+static inline void remove_record(struct page *page, struct pevent_record *record) {}
+static inline void add_record(struct page *page, struct pevent_record *record) {}
+static const char *show_records(struct list_head *pages)
+{
+	return "";
+}
+#endif
+
 static int init_cpu(struct tracecmd_input *handle, int cpu);
 
 static int do_read(struct tracecmd_input *handle, void *data, int size)
@@ -288,10 +352,12 @@ static int read_ftrace_file(struct tracecmd_input *handle,
 		return -1;
 	}
 
-	if (print)
+	if (print) {
 		printf("%.*s\n", (int)size, buf);
-	else
-		pevent_parse_event(pevent, buf, size, "ftrace");
+	} else {
+		if (pevent_parse_event(pevent, buf, size, "ftrace"))
+			pevent->parsing_failures = 1;
+	}
 	free(buf);
 
 	return 0;
@@ -313,10 +379,12 @@ static int read_event_file(struct tracecmd_input *handle,
 		return -1;
 	}
 
-	if (print)
+	if (print) {
 		printf("%.*s\n", (int)size, buf);
-	else
-		pevent_parse_event(pevent, buf, size, system);
+	} else {
+		if (pevent_parse_event(pevent, buf, size, system))
+			pevent->parsing_failures = 1;
+	}
 	free(buf);
 
 	return 0;
@@ -432,7 +500,7 @@ static int read_ftrace_printk(struct tracecmd_input *handle)
 	if (size < 0)
 		return -1;
 
-	buf = malloc(size);
+	buf = malloc(size + 1);
 	if (!buf)
 		return -1;
 	if (do_read_check(handle, buf, size)) {
@@ -440,6 +508,8 @@ static int read_ftrace_printk(struct tracecmd_input *handle)
 		return -1;
 	}
 
+	buf[size] = 0;
+
 	parse_ftrace_printk(handle->pevent, buf, size);
 
 	free(buf);
@@ -479,67 +549,9 @@ int tracecmd_read_headers(struct tracecmd_input *handle)
 	if (ret < 0)
 		return -1;
 
-	/* register default ftrace functions first */
-	tracecmd_ftrace_overrides(handle);
-
 	return 0;
 }
 
-static unsigned int type4host(struct tracecmd_input *handle,
-			      unsigned int type_len_ts)
-{
-	struct pevent *pevent = handle->pevent;
-
-	if (pevent->file_bigendian)
-		return (type_len_ts >> 29) & 3;
-	else
-		return type_len_ts & 3;
-}
-
-static unsigned int len4host(struct tracecmd_input *handle,
-			     unsigned int type_len_ts)
-{
-	struct pevent *pevent = handle->pevent;
-
-	if (pevent->file_bigendian)
-		return (type_len_ts >> 27) & 7;
-	else
-		return (type_len_ts >> 2) & 7;
-}
-
-static unsigned int type_len4host(struct tracecmd_input *handle,
-				  unsigned int type_len_ts)
-{
-	struct pevent *pevent = handle->pevent;
-
-	if (pevent->file_bigendian)
-		return (type_len_ts >> 27) & ((1 << 5) - 1);
-	else
-		return type_len_ts & ((1 << 5) - 1);
-}
-
-static unsigned int ts4host(struct tracecmd_input *handle,
-			    unsigned int type_len_ts)
-{
-	struct pevent *pevent = handle->pevent;
-
-	if (pevent->file_bigendian)
-		return type_len_ts & ((1 << 27) - 1);
-	else
-		return type_len_ts >> 5;
-}
-
-static unsigned int read_type_len_ts(struct tracecmd_input *handle, void *ptr)
-{
-	return data2host4(handle->pevent, ptr);
-}
-
-static int calc_index(struct tracecmd_input *handle,
-		      void *ptr, int cpu)
-{
-	return (unsigned long)ptr - (unsigned long)handle->cpu_data[cpu].page->map;
-}
-
 static unsigned long long calc_page_offset(struct tracecmd_input *handle,
 					   unsigned long long offset)
 {
@@ -575,7 +587,7 @@ static struct page *allocate_page(struct tracecmd_input *handle,
 	struct page *page;
 	int ret;
 
-	list_for_each_entry(page, &cpu_data->pages, struct page, list) {
+	list_for_each_entry(page, &cpu_data->pages, list) {
 		if (page->offset == offset) {
 			page->ref_count++;
 			return page;
@@ -637,7 +649,8 @@ static void __free_page(struct tracecmd_input *handle, struct page *page)
 
 static void free_page(struct tracecmd_input *handle, int cpu)
 {
-	if (!handle->cpu_data[cpu].page)
+	if (!handle->cpu_data || cpu >= handle->cpus ||
+	    !handle->cpu_data[cpu].page)
 		return;
 
 	__free_page(handle, handle->cpu_data[cpu].page);
@@ -645,37 +658,62 @@ static void free_page(struct tracecmd_input *handle, int cpu)
 	handle->cpu_data[cpu].page = NULL;
 }
 
-static void __free_record(struct record *record)
+static void __free_record(struct pevent_record *record)
 {
-	if (record->private) {
-		struct page *page = record->private;
+	if (record->priv) {
+		struct page *page = record->priv;
+		remove_record(page, record);
 		__free_page(page->handle, page);
 	}
 
 	free(record);
 }
 
-void free_record(struct record *record)
+void free_record(struct pevent_record *record)
 {
 	if (!record)
 		return;
 
+	if (!record->ref_count)
+		die("record ref count is zero!");
+
+	record->ref_count--;
+
+	if (record->ref_count)
+		return;
+
 	if (record->locked)
 		die("freeing record when it is locked!");
+
 	record->data = NULL;
 
 	__free_record(record);
 }
 
+void tracecmd_record_ref(struct pevent_record *record)
+{
+	record->ref_count++;
+#if DEBUG_RECORD
+	/* Update locating of last reference */
+	record->alloc_addr = (unsigned long)__builtin_return_address(0);
+#endif
+}
+
 static void free_next(struct tracecmd_input *handle, int cpu)
 {
-	struct record *record = handle->cpu_data[cpu].next;
+	struct pevent_record *record;
 
+	if (!handle->cpu_data || cpu >= handle->cpus)
+		return;
+
+	record = handle->cpu_data[cpu].next;
 	if (!record)
 		return;
 
 	handle->cpu_data[cpu].next = NULL;
-	__free_record(record);
+
+	record->locked = 0;
+	free_record(record);
 }
 
 /*
@@ -685,7 +723,7 @@ static int update_page_info(struct tracecmd_input *handle, int cpu)
 {
 	struct pevent *pevent = handle->pevent;
 	void *ptr = handle->cpu_data[cpu].page->map;
-	unsigned int flags;
+	struct kbuffer *kbuf = handle->cpu_data[cpu].kbuf;
 
 	/* FIXME: handle header page */
 	if (pevent->header_page_ts_size != 8) {
@@ -693,43 +731,8 @@ static int update_page_info(struct tracecmd_input *handle, int cpu)
 		return -1;
 	}
 
-	handle->cpu_data[cpu].timestamp = data2host8(pevent, ptr);
-	ptr += 8;
-	switch (pevent->header_page_size_size) {
-	case 4:
-		flags = data2host4(pevent, ptr);
-		ptr += 4;
-		break;
-	case 8:
-		flags = (unsigned int)data2host8(pevent, ptr);
-		ptr += 8;
-		break;
-	default:
-		warning("bad long size");
-		return -1;
-	}
-
-	handle->cpu_data[cpu].page_size = flags & COMMIT_MASK;
-
-	if (flags & MISSING_EVENTS) {
-		if (flags & MISSING_STORED) {
-			ptr += handle->cpu_data[cpu].page_size;
-			switch (pevent->header_page_size_size) {
-			case 4:
-				handle->cpu_data[cpu].page->lost_events =
-					data2host4(pevent, ptr);
-				break;
-			case 8:
-				handle->cpu_data[cpu].page->lost_events =
-					data2host8(pevent, ptr);
-				break;
-			}
-		} else
-			handle->cpu_data[cpu].page->lost_events = -1;
-	} else
-		handle->cpu_data[cpu].page->lost_events = 0;
-
-	handle->cpu_data[cpu].index = 0;
+	kbuffer_load_subbuffer(kbuf, ptr);
+	handle->cpu_data[cpu].timestamp = kbuffer_timestamp(kbuf) + handle->ts_offset;
 
 	return 0;
 }
@@ -768,8 +771,6 @@ static int get_page(struct tracecmd_input *handle, int cpu,
 	}
 
 	handle->cpu_data[cpu].offset = offset;
-	handle->cpu_data[cpu].timestamp = 0;
-	handle->cpu_data[cpu].index = 0;
 	handle->cpu_data[cpu].size = (handle->cpu_data[cpu].file_offset +
 				      handle->cpu_data[cpu].file_size) -
 					offset;
@@ -785,6 +786,7 @@ static int get_page(struct tracecmd_input *handle, int cpu,
 
 	return 0;
 }
+
 static int get_next_page(struct tracecmd_input *handle, int cpu)
 {
 	off64_t offset;
@@ -796,7 +798,6 @@ static int get_next_page(struct tracecmd_input *handle, int cpu)
 
 	if (handle->cpu_data[cpu].size <= handle->page_size) {
 		handle->cpu_data[cpu].offset = 0;
-		handle->cpu_data[cpu].timestamp = 0;
 		return 0;
 	}
 
@@ -805,92 +806,11 @@ static int get_next_page(struct tracecmd_input *handle, int cpu)
 	return get_page(handle, cpu, offset);
 }
 
-enum old_ring_buffer_type {
-	OLD_RINGBUF_TYPE_PADDING,
-	OLD_RINGBUF_TYPE_TIME_EXTEND,
-	OLD_RINGBUF_TYPE_TIME_STAMP,
-	OLD_RINGBUF_TYPE_DATA,
-};
-
-static struct record *
-read_old_format(struct tracecmd_input *handle, void **ptr, int cpu)
-{
-	struct pevent *pevent = handle->pevent;
-	struct record *data;
-	unsigned long long extend;
-	unsigned int type_len_ts;
-	unsigned int type;
-	unsigned int len;
-	unsigned int delta;
-	unsigned int length;
-	int index;
-
-	index = calc_index(handle, *ptr, cpu);
-
-	type_len_ts = read_type_len_ts(handle, *ptr);
-	*ptr += 4;
-
-	type = type4host(handle, type_len_ts);
-	len = len4host(handle, type_len_ts);
-	delta = ts4host(handle, type_len_ts);
-
-	switch (type) {
-	case OLD_RINGBUF_TYPE_PADDING:
-		*ptr = (void *)(((unsigned long)*ptr + (handle->page_size - 1)) &
-				~(handle->page_size - 1));
-		return NULL;
-
-	case OLD_RINGBUF_TYPE_TIME_EXTEND:
-		extend = data2host4(pevent, ptr);
-		extend <<= TS_SHIFT;
-		extend += delta;
-		handle->cpu_data[cpu].timestamp += extend;
-		*ptr += 4;
-		return NULL;
-
-	case OLD_RINGBUF_TYPE_TIME_STAMP:
-		warning("should not be here");
-		return NULL;
-		break;
-	default:
-		if (len)
-			length = len * 4;
-		else {
-			length = data2host4(pevent, *ptr);
-			length -= 4;
-			*ptr += 4;
-		}
-		break;
-	}
-
-	handle->cpu_data[cpu].timestamp += delta;
-
-	data = malloc(sizeof(*data));
-	if (!data)
-		return NULL;
-	memset(data, 0, sizeof(*data));
-
-	data->ts = handle->cpu_data[cpu].timestamp;
-	data->size = length;
-	data->data = *ptr;
-	data->offset = handle->cpu_data[cpu].offset + index;
-
-
-	*ptr += ((length+3)/4) * 4;
-
-	handle->cpu_data[cpu].index = calc_index(handle, *ptr, cpu);
-	handle->cpu_data[cpu].next = data;
-
-	data->record_size = handle->cpu_data[cpu].index - index;
-
-	return data;
-}
-
-static struct record *
+static struct pevent_record *
 peek_event(struct tracecmd_input *handle, unsigned long long offset,
 	   int cpu)
 {
-	struct record *record = NULL;
+	struct pevent_record *record = NULL;
 
 	/*
 	 * Since the timestamp is calculated from the beginning
@@ -899,24 +819,22 @@ peek_event(struct tracecmd_input *handle, unsigned long long offset,
 	 * tracecmd_read_at.
 	 */
 	update_page_info(handle, cpu);
-	free_next(handle, cpu);
 
 	do {
-		free_record(record);
+		free_next(handle, cpu);
 		record = tracecmd_peek_data(handle, cpu);
 		if (record && (record->offset + record->record_size) > offset)
 			break;
-		record = tracecmd_read_data(handle, cpu);
         } while (record);
 
 	return record;
 }
 
-static struct record *
+static struct pevent_record *
 read_event(struct tracecmd_input *handle, unsigned long long offset,
 	   int cpu)
 {
-	struct record *record;
+	struct pevent_record *record;
 
 	record = peek_event(handle, offset, cpu);
 	if (record)
@@ -924,7 +842,7 @@ read_event(struct tracecmd_input *handle, unsigned long long offset,
 	return record;
 }
 
-static struct record *
+static struct pevent_record *
 find_and_peek_event(struct tracecmd_input *handle, unsigned long long offset,
 		    int *pcpu)
 {
@@ -956,11 +874,11 @@ find_and_peek_event(struct tracecmd_input *handle, unsigned long long offset,
 }
 
 
-static struct record *
+static struct pevent_record *
 find_and_read_event(struct tracecmd_input *handle, unsigned long long offset,
 		    int *pcpu)
 {
-	struct record *record;
+	struct pevent_record *record;
 	int cpu;
 
 	record = find_and_peek_event(handle, offset, &cpu);
@@ -985,7 +903,7 @@ find_and_read_event(struct tracecmd_input *handle, unsigned long long offset,
  *
  * The record returned must be freed.
  */
-struct record *
+struct pevent_record *
 tracecmd_read_at(struct tracecmd_input *handle, unsigned long long offset,
 		 int *pcpu)
 {
@@ -1027,20 +945,18 @@ tracecmd_read_at(struct tracecmd_input *handle, unsigned long long offset,
  *        -1 on error.
  */
 int tracecmd_refresh_record(struct tracecmd_input *handle,
-			    struct record *record)
+			    struct pevent_record *record)
 {
 	unsigned long long page_offset;
 	int cpu = record->cpu;
 	struct cpu_data *cpu_data = &handle->cpu_data[cpu];
-	unsigned int type_len_ts;
-	unsigned int len;
 	int index;
 	int ret;
 
 	page_offset = calc_page_offset(handle, record->offset);
 	index = record->offset & (handle->page_size - 1);
 
-	ret =get_page(handle, record->cpu, page_offset);
+	ret = get_page(handle, record->cpu, page_offset);
 	if (ret < 0)
 		return -1;
 
@@ -1048,16 +964,7 @@ int tracecmd_refresh_record(struct tracecmd_input *handle,
 	if (ret)
 		return 1;
 
-	record->data = cpu_data->page->map + index;
-
-	type_len_ts = read_type_len_ts(handle, record->data);
-	len = len4host(handle, type_len_ts);
-
-	/* The data starts either 4 or 8 bytes from offset */
-	record->data += len ? 4 : 8;
-
-	/* The get_page resets the index, set the index after this record */
-	cpu_data->index = index + record->record_size;
+	record->data = kbuffer_read_at_offset(cpu_data->kbuf, index, &record->ts);
 	cpu_data->timestamp = record->ts;
 
 	return 0;
@@ -1072,13 +979,19 @@ int tracecmd_refresh_record(struct tracecmd_input *handle,
  *
  * The record returned must be freed.
  */
-struct record *
+struct pevent_record *
 tracecmd_read_cpu_first(struct tracecmd_input *handle, int cpu)
 {
-	if (get_page(handle, cpu, handle->cpu_data[cpu].file_offset) < 0)
+	int ret;
+
+	ret = get_page(handle, cpu, handle->cpu_data[cpu].file_offset);
+	if (ret < 0)
 		return NULL;
 
-	handle->cpu_data[cpu].index = 0;
+	/* If the page was already mapped, we need to reset it */
+	if (ret)
+		update_page_info(handle, cpu);
+		
 	free_next(handle, cpu);
 
 	return tracecmd_read_data(handle, cpu);
@@ -1093,10 +1006,10 @@ tracecmd_read_cpu_first(struct tracecmd_input *handle, int cpu)
  *
  * The record returned must be freed.
  */
-struct record *
+struct pevent_record *
 tracecmd_read_cpu_last(struct tracecmd_input *handle, int cpu)
 {
-	struct record *record = NULL;
+	struct pevent_record *record = NULL;
 	off64_t offset, page_offset;
 
 	offset = handle->cpu_data[cpu].file_offset +
@@ -1233,8 +1146,6 @@ tracecmd_set_cpu_to_timestamp(struct tracecmd_input *handle, int cpu,
 	    cpu_data->offset > cpu_data->file_offset)
 		get_page(handle, cpu, cpu_data->offset - handle->page_size);
 
-	cpu_data->index = 0;
-
 	return 0;
 }
 
@@ -1317,8 +1228,7 @@ unsigned long long
 tracecmd_get_cursor(struct tracecmd_input *handle, int cpu)
 {
 	struct cpu_data *cpu_data = &handle->cpu_data[cpu];
-	struct pevent *pevent;
-	int index;
+	struct kbuffer *kbuf = cpu_data->kbuf;
 
 	if (cpu < 0 || cpu >= handle->cpus)
 		return 0;
@@ -1342,59 +1252,7 @@ tracecmd_get_cursor(struct tracecmd_input *handle, int cpu)
 	    cpu_data->file_size)
 		return cpu_data->offset;
 
-	pevent = handle->pevent;
-
-	index = cpu_data->index ? cpu_data->index :
-		pevent->header_page_data_offset;
-
-	return cpu_data->offset + index;
-}
-
-
-static unsigned int
-translate_data(struct tracecmd_input *handle,
-	       void **ptr, unsigned long long *delta, int *length)
-{
-	struct pevent *pevent = handle->pevent;
-	unsigned long long extend;
-	unsigned int type_len_ts;
-	unsigned int type_len;
-
-	type_len_ts = read_type_len_ts(handle, *ptr);
-	*ptr += 4;
-
-	type_len = type_len4host(handle, type_len_ts);
-	*delta = ts4host(handle, type_len_ts);
-
-	switch (type_len) {
-	case RINGBUF_TYPE_PADDING:
-		*length = data2host4(pevent, *ptr);
-		*length *= 4;
-		*ptr += *length;
-		break;
-
-	case RINGBUF_TYPE_TIME_EXTEND:
-		extend = data2host4(pevent, *ptr);
-		*ptr += 4;
-		extend <<= TS_SHIFT;
-		extend += *delta;
-		*delta = extend;
-		break;
-
-	case RINGBUF_TYPE_TIME_STAMP:
-		*ptr += 12;
-		break;
-	case 0:
-		*length = data2host4(pevent, *ptr) - 4;
-		*length = (*length + 3) & ~3;
-		*ptr += 4;
-		break;
-	default:
-		*length = type_len * 4;
-		break;
-	}
-
-	return type_len;
+	return cpu_data->offset + kbuffer_curr_offset(kbuf);
 }
 
 /**
@@ -1413,12 +1271,14 @@ translate_data(struct tracecmd_input *handle,
  *
  * The record returned must be freed.
  */
-struct record *
+struct pevent_record *
 tracecmd_translate_data(struct tracecmd_input *handle,
 			void *ptr, int size)
 {
-	struct record *record;
-	unsigned int type_len;
+	struct pevent *pevent = handle->pevent;
+	struct pevent_record *record;
+	unsigned int length;
+	int swap = 1;
 
 	/* minimum record read is 8, (warn?) (TODO: make 8 into macro) */
 	if (size < 8)
@@ -1429,18 +1289,105 @@ tracecmd_translate_data(struct tracecmd_input *handle,
 		return NULL;
 	memset(record, 0, sizeof(*record));
 
-	record->data = ptr;
-	type_len = translate_data(handle, &record->data, &record->ts, &record->size);
-	switch (type_len) {
-	case RINGBUF_TYPE_PADDING:
-	case RINGBUF_TYPE_TIME_EXTEND:
-	case RINGBUF_TYPE_TIME_STAMP:
-		record->data = NULL;
-		break;
-	default:
-		break;
+	record->ref_count = 1;
+	if (pevent->host_bigendian == pevent->file_bigendian)
+		swap = 0;
+	record->data = kbuffer_translate_data(swap, ptr, &length);
+	record->size = length;
+	if (record->data)
+		record->record_size = record->size + (record->data - ptr);
+
+	return record;
+}
+
+/**
+ * tracecmd_read_page_record - read a record off of a page
+ * @pevent: pevent used to parse the page
+ * @page: the page to read
+ * @size: the size of the page
+ * @last_record: last record read from this page.
+ *
+ * If a ring buffer page is available, and the need to parse it
+ * without having a handle, then this function can be used.
+ *
+ * The @pevent needs to be initialized to have the page header information
+ * already available.
+ *
+ * The @last_record is used to know where to read the next record from.
+ * If @last_record is NULL, the first record on the page will be read.
+ *
+ * Returns:
+ *  A newly allocated record that must be freed with free_record() if
+ *  a record is found. Otherwise NULL is returned if the record is bad
+ *  or no more records exist.
+ */
+struct pevent_record *
+tracecmd_read_page_record(struct pevent *pevent, void *page, int size,
+			  struct pevent_record *last_record)
+{
+	unsigned long long ts;
+	struct kbuffer *kbuf;
+	struct pevent_record *record = NULL;
+	enum kbuffer_long_size long_size;
+	enum kbuffer_endian endian;
+	void *ptr;
+
+	if (pevent->file_bigendian)
+		endian = KBUFFER_ENDIAN_BIG;
+	else
+		endian = KBUFFER_ENDIAN_LITTLE;
+
+	if (pevent->header_page_size_size == 8)
+		long_size = KBUFFER_LSIZE_8;
+	else
+		long_size = KBUFFER_LSIZE_4;
+
+	kbuf = kbuffer_alloc(long_size, endian);
+	if (!kbuf)
+		return NULL;
+
+	kbuffer_load_subbuffer(kbuf, page);
+	if (kbuffer_subbuffer_size(kbuf) > size) {
+		warning("tracecmd_read_page_record: page_size > size");
+		goto out_free;
 	}
 
+	if (last_record) {
+		if (last_record->data < page || last_record->data >= (page + size)) {
+			warning("tracecmd_read_page_record: bad last record (size=%u)",
+				last_record->size);
+			goto out_free;
+		}
+
+		do {
+			ptr = kbuffer_next_event(kbuf, NULL);
+			if (!ptr)
+				break;
+		} while (ptr < last_record->data);
+		if (ptr != last_record->data) {
+			warning("tracecmd_read_page_record: could not find last_record");
+			goto out_free;
+		}
+	}
+
+	ptr = kbuffer_read_event(kbuf, &ts);
+	if (!ptr)
+		goto out_free;
+
+	record = malloc(sizeof(*record));
+	if (!record)
+		return NULL;
+	memset(record, 0, sizeof(*record));
+
+	record->ts = ts;
+	record->size = kbuffer_event_size(kbuf);
+	record->record_size = kbuffer_curr_size(kbuf);
+	record->cpu = 0;
+	record->data = ptr;
+	record->ref_count = 1;
+
+ out_free:
+	kbuffer_free(kbuf);
 	return record;
 }
 
@@ -1451,24 +1398,22 @@ tracecmd_translate_data(struct tracecmd_input *handle,
  *
  * This returns the record at the current location of the CPU
  * iterator. It does not increment the CPU iterator.
- *
- * NOTE: Do not free the record returned, it is stored in the @handle.
  */
-struct record *
+struct pevent_record *
 tracecmd_peek_data(struct tracecmd_input *handle, int cpu)
 {
-	struct pevent *pevent = handle->pevent;
-	struct record *record;
-	struct page *page = handle->cpu_data[cpu].page;
-	int index = handle->cpu_data[cpu].index;
-	void *ptr;
-	unsigned long long extend;
-	unsigned int type_len;
-	long long missed_events = 0;
-	int length;
+	struct pevent_record *record;
+	unsigned long long ts;
+	struct kbuffer *kbuf;
+	struct page *page;
+	int index;
+	void *data;
 
-	if (index < 0)
-		die("negative index on cpu iterator %d", cpu);
+	if (cpu >= handle->cpus)
+		return NULL;
+
+	page = handle->cpu_data[cpu].page;
+	kbuf = handle->cpu_data[cpu].kbuf;
 
 	/* Hack to work around function graph read ahead */
 	tracecmd_curr_thread_handle = handle;
@@ -1489,58 +1434,21 @@ tracecmd_peek_data(struct tracecmd_input *handle, int cpu)
 		free_next(handle, cpu);
 	}
 
+read_again:
 	if (!page)
 		return NULL;
 
-	ptr = page->map + index;
-
-	if (!index) {
-		missed_events = page->lost_events;
-		ptr = handle->cpu_data[cpu].page->map + pevent->header_page_data_offset;
-	}
-
-read_again:
-	index = calc_index(handle, ptr, cpu);
-
-	if (index < 0)
-		die("negative index on cpu record %d", cpu);
-
-	if (index >= handle->cpu_data[cpu].page_size + pevent->header_page_data_offset) {
+	data = kbuffer_read_event(kbuf, &ts);
+	if (!data) {
 		if (get_next_page(handle, cpu))
 			return NULL;
-		return tracecmd_peek_data(handle, cpu);
-	}
-
-	if (pevent->old_format) {
-		record = read_old_format(handle, &ptr, cpu);
-		if (!record) {
-			if (!ptr)
-				return NULL;
-			goto read_again;
-		}
-		record->cpu = cpu;
-		return record;
-	}
-
-	type_len = translate_data(handle, &ptr, &extend, &length);
-
-	switch (type_len) {
-	case RINGBUF_TYPE_PADDING:
-		if (!extend) {
-			warning("error, hit unexpected end of page");
-			return NULL;
-		}
-		/* fall through */
-	case RINGBUF_TYPE_TIME_EXTEND:
-		handle->cpu_data[cpu].timestamp += extend;
-		/* fall through */
-	case RINGBUF_TYPE_TIME_STAMP:
+		page = handle->cpu_data[cpu].page;
 		goto read_again;
-	default:
-		break;
 	}
 
-	handle->cpu_data[cpu].timestamp += extend;
+	handle->cpu_data[cpu].timestamp = ts + handle->ts_offset;
+
+	index = kbuffer_curr_offset(kbuf);
 
 	record = malloc(sizeof(*record));
 	if (!record)
@@ -1548,22 +1456,23 @@ read_again:
 	memset(record, 0, sizeof(*record));
 
 	record->ts = handle->cpu_data[cpu].timestamp;
-	record->size = length;
+	record->size = kbuffer_event_size(kbuf);
 	record->cpu = cpu;
-	record->data = ptr;
+	record->data = data;
 	record->offset = handle->cpu_data[cpu].offset + index;
-	record->missed_events = missed_events;
+	record->missed_events = kbuffer_missed_events(kbuf);
+	record->ref_count = 1;
+	record->locked = 1;
 
-	ptr += length;
-
-	handle->cpu_data[cpu].index = calc_index(handle, ptr, cpu);
 	handle->cpu_data[cpu].next = record;
 
-	record->record_size = handle->cpu_data[cpu].index - index;
-	record->private = page;
-	record->locked = 1;
+	record->record_size = kbuffer_curr_size(kbuf);
+	record->priv = page;
+	add_record(page, record);
 	page->ref_count++;
 
+	kbuffer_next_event(kbuf, NULL);
+
 	return record;
 }
 
@@ -1577,16 +1486,19 @@ read_again:
  *
  * The record returned must be freed.
  */
-struct record *
+struct pevent_record *
 tracecmd_read_data(struct tracecmd_input *handle, int cpu)
 {
-	struct record *record;
+	struct pevent_record *record;
 
 	record = tracecmd_peek_data(handle, cpu);
 	handle->cpu_data[cpu].next = NULL;
-	if (record)
+	if (record) {
 		record->locked = 0;
-
+#if DEBUG_RECORD
+		record->alloc_addr = (unsigned long)__builtin_return_address(0);
+#endif
+	}
 	return record;
 }
 
@@ -1607,11 +1519,12 @@ tracecmd_read_data(struct tracecmd_input *handle, int cpu)
  *
  * The record returned must be freed.
  */
-struct record *
+struct pevent_record *
 tracecmd_read_next_data(struct tracecmd_input *handle, int *rec_cpu)
 {
 	unsigned long long ts;
-	struct record *record;
+	struct pevent_record *record;
+	int first_record = 1;
 	int next;
 	int cpu;
 
@@ -1623,9 +1536,10 @@ tracecmd_read_next_data(struct tracecmd_input *handle, int *rec_cpu)
 
 	for (cpu = 0; cpu < handle->cpus; cpu++) {
 		record = tracecmd_peek_data(handle, cpu);
-		if (record && (!ts || record->ts < ts)) {
+		if (record && (first_record || record->ts < ts)) {
 			ts = record->ts;
 			next = cpu;
+			first_record = 0;
 		}
 	}
 
@@ -1647,17 +1561,16 @@ tracecmd_read_next_data(struct tracecmd_input *handle, int *rec_cpu)
  * @record is the first record, NULL is returned. The cursor is set
  * as if the previous record was read by tracecmd_read_data().
  *
- * @record can not be NULL, otherwise NULL is returned.
+ * @record can not be NULL, otherwise NULL is returned; the
+ * record ownership goes to this function.
  *
  * Note, this is not that fast of an algorithm, since it needs
  * to build the timestamp for the record.
  *
- * Note 2: This may free any record allocoted with tracecmd_peek_data().
- *
  * The record returned must be freed with free_record().
  */
-struct record *
-tracecmd_read_prev(struct tracecmd_input *handle, struct record *record)
+struct pevent_record *
+tracecmd_read_prev(struct tracecmd_input *handle, struct pevent_record *record)
 {
 	unsigned long long offset, page_offset;;
 	struct cpu_data *cpu_data;
@@ -1702,6 +1615,9 @@ tracecmd_read_prev(struct tracecmd_input *handle, struct record *record)
 		/* we found our record */
 		return tracecmd_read_at(handle, page_offset + index, NULL);
 
+	/* reset the index to start at the beginning of the page */
+	update_page_info(handle, cpu);
+
 	/* The previous record is on the previous page */
 	for (;;) {
 		/* check if this is the first page */
@@ -1782,30 +1698,62 @@ static int init_cpu(struct tracecmd_input *handle, int cpu)
 
 static int handle_options(struct tracecmd_input *handle)
 {
+	unsigned long long offset;
 	unsigned short option;
 	unsigned int size;
+	char *cpustats = NULL;
+	unsigned int cpustats_size = 0;
 	char *buf;
 
-	do {
+	for (;;) {
 		if (do_read_check(handle, &option, 2))
 			return -1;
 
+		if (option == TRACECMD_OPTION_DONE)
+			break;
+
+		/* next 4 bytes is the size of the option */
+		if (do_read_check(handle, &size, 4))
+			return -1;
+		size = __data2host4(handle->pevent, size);
+		buf = malloc_or_die(size);
+		if (do_read_check(handle, buf, size))
+			return -1;
+
 		switch (option) {
-		case TRACECMD_OPTION_DONE:
+		case TRACECMD_OPTION_DATE:
+			/*
+			 * A time has been mapped that is the
+			 * difference between the timestamps and
+			 * gtod. It is stored as ASCII with '0x'
+			 * appended.
+			 */
+			if (handle->flags & TRACECMD_FL_IGNORE_DATE)
+				break;
+			offset = strtoll(buf, NULL, 0);
+			/* Convert from micro to nano */
+			offset *= 1000;
+			handle->ts_offset = offset;
+			break;
+		case TRACECMD_OPTION_CPUSTAT:
+			buf[size-1] = '\n';
+			cpustats = realloc(cpustats, cpustats_size + size + 1);
+			if (!cpustats)
+				die("realloc");
+			memcpy(cpustats + cpustats_size, buf, size);
+			cpustats_size += size;
+			cpustats[cpustats_size] = 0;
 			break;
 		default:
 			warning("unknown option %d", option);
-			/* next 4 bytes is the size of the option */
-			if (do_read_check(handle, &size, 4))
-				return -1;
-			size = __data2host4(handle->pevent, size);
-			buf = malloc_or_die(size);
-			if (do_read_check(handle, buf, size))
-				return -1;
-			free(buf);
 			break;
 		}
-	} while (option != TRACECMD_OPTION_DONE);
+
+		free(buf);
+
+	}
+
+	handle->cpustats = cpustats;
 
 	return 0;
 }
@@ -1820,6 +1768,8 @@ static int handle_options(struct tracecmd_input *handle)
 int tracecmd_init_data(struct tracecmd_input *handle)
 {
 	struct pevent *pevent = handle->pevent;
+	enum kbuffer_long_size long_size;
+	enum kbuffer_endian endian;
 	unsigned long long size;
 	char *cmdlines;
 	char buf[10];
@@ -1875,22 +1825,55 @@ int tracecmd_init_data(struct tracecmd_input *handle)
 	if (force_read)
 		handle->read_page = 1;
 
+	if (handle->long_size == 8)
+		long_size = KBUFFER_LSIZE_8;
+	else
+		long_size = KBUFFER_LSIZE_4;
+
+	if (handle->pevent->file_bigendian)
+		endian = KBUFFER_ENDIAN_BIG;
+	else
+		endian = KBUFFER_ENDIAN_LITTLE;
+
 	for (cpu = 0; cpu < handle->cpus; cpu++) {
 		unsigned long long offset;
 
 		handle->cpu_data[cpu].cpu = cpu;
 
+		handle->cpu_data[cpu].kbuf = kbuffer_alloc(long_size, endian);
+		if (!handle->cpu_data[cpu].kbuf)
+			goto out_free;
+		if (pevent->old_format)
+			kbuffer_set_old_format(handle->cpu_data[cpu].kbuf);
+
 		offset = read8(handle);
 		size = read8(handle);
 
 		handle->cpu_data[cpu].file_offset = offset;
 		handle->cpu_data[cpu].file_size = size;
 
+		if (size && (offset + size > handle->total_file_size)) {
+			/* this happens if the file got truncated */
+			printf("File possibly truncated. "
+				"Need at least %llu, but file size is %zu.\n",
+				offset + size, handle->total_file_size);
+			errno = EINVAL;
+			goto out_free;
+		}
+
 		if (init_cpu(handle, cpu))
-			return -1;
+			goto out_free;
 	}
 
+	tracecmd_blk_hack(handle);
+
 	return 0;
+out_free:
+	for ( ; cpu >= 0; cpu--) {
+		free_page(handle, cpu);
+		kbuffer_free(handle->cpu_data[cpu].kbuf);
+	}
+	return -1;
 }
 
 /**
@@ -1916,6 +1899,37 @@ void tracecmd_print_events(struct tracecmd_input *handle)
 	return;
 }
 
+/* Show the cpu data stats */
+static void show_cpu_stats(struct tracecmd_input *handle)
+{
+	struct cpu_data *cpu_data;
+	int i;
+
+	for (i = 0; i < handle->cpus; i++) {
+		cpu_data = &handle->cpu_data[i];
+		printf("CPU%d data recorded at offset=0x%llx\n",
+		       i, cpu_data->file_offset);
+		printf("    %lld bytes in size\n", cpu_data->file_size);
+	}
+}
+
+/**
+ * tracecmd_print_stats - prints the stats recorded in the options.
+ * @handle: input handle for the trace.dat file
+ *
+ * Looks for the option TRACECMD_OPTION_CPUSTAT and prints out what's
+ * stored there, if it is found. Otherwise it prints that none were found.
+ */
+void tracecmd_print_stats(struct tracecmd_input *handle)
+{
+	if (handle->cpustats)
+		printf("%s\n", handle->cpustats);
+	else
+		printf(" No stats in this file\n");
+
+	show_cpu_stats(handle);
+}
+
 /**
  * tracecmd_alloc_fd - create a tracecmd_input handle from a file descriptor
  * @fd: the file descriptor for the trace.dat file
@@ -1969,6 +1983,9 @@ struct tracecmd_input *tracecmd_alloc_fd(int fd)
 	if (!handle->pevent)
 		goto failed_read;
 
+	/* register default ftrace functions first */
+	tracecmd_ftrace_overrides(handle, &handle->finfo);
+
 	handle->plugin_list = tracecmd_load_plugins(handle->pevent);
 
 	handle->pevent->file_bigendian = buf[0];
@@ -1982,6 +1999,12 @@ struct tracecmd_input *tracecmd_alloc_fd(int fd)
 	handle->header_files_start =
 		lseek64(handle->fd, 0, SEEK_CUR);
 
+	handle->total_file_size =
+		lseek64(handle->fd, 0, SEEK_END);
+
+	handle->header_files_start =
+		lseek64(handle->fd, handle->header_files_start, SEEK_SET);
+
 	return handle;
 
  failed_read:
@@ -2097,8 +2120,13 @@ void tracecmd_close(struct tracecmd_input *handle)
 		/* The tracecmd_peek_data may have cached a record */
 		free_next(handle, cpu);
 		free_page(handle, cpu);
-		if (!list_empty(&handle->cpu_data[cpu].pages))
-			warning("pages still allocated on cpu %d", cpu);
+		if (handle->cpu_data) {
+			kbuffer_free(handle->cpu_data[cpu].kbuf);
+
+			if (!list_empty(&handle->cpu_data[cpu].pages))
+				warning("pages still allocated on cpu %d%s",
+					cpu, show_records(&handle->cpu_data[cpu].pages));
+		}
 	}
 
 	free(handle->cpu_data);
