@@ -13,8 +13,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not,  see <http://www.gnu.org/licenses>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
@@ -33,13 +32,25 @@ void usage(char **argv);
 extern int silence_warnings;
 extern int show_status;
 
-struct tracecmd_input *read_trace_header(void);
+struct tracecmd_input *read_trace_header(const char *file);
 int read_trace_files(void);
 
+void trace_record(int argc, char **argv);
+
 void trace_report(int argc, char **argv);
 
 void trace_split(int argc, char **argv);
 
 void trace_listen(int argc, char **argv);
 
+void trace_restore(int argc, char **argv);
+
+void trace_stack(int argc, char **argv);
+
+void trace_option(int argc, char **argv);
+
+void trace_hist(int argc, char **argv);
+
+void trace_mem(int argc, char **argv);
+
 #endif /* __TRACE_LOCAL_H */
