@@ -13,13 +13,11 @@
  * GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+ * License along with this program; if not,  see <http://www.gnu.org/licenses>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
 #define _LARGEFILE64_SOURCE
-#define _GNU_SOURCE
 #include <dirent.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -35,24 +33,63 @@
 #include <unistd.h>
 #include <ctype.h>
 #include <errno.h>
+#include <glob.h>
 
 #include "trace-cmd-local.h"
 #include "version.h"
 
+static struct tracecmd_event_list all_event_list = {
+	.next = NULL,
+	.glob = "all"
+};
+
+struct tracecmd_option {
+	unsigned short	id;
+	int		size;
+	void		*data;
+};
+
 struct tracecmd_output {
 	int		fd;
 	int		page_size;
 	int		cpus;
 	struct pevent	*pevent;
 	char		*tracing_dir;
+	int		options_written;
+	int		nr_options;
+	struct tracecmd_option *options;
 };
 
-static int
-do_write_check(struct tracecmd_output *handle, void *data, int size)
+struct list_event {
+	struct list_event		*next;
+	char				*name;
+	char				*file;
+};
+
+struct list_event_system {
+	struct list_event_system	*next;
+	struct list_event		*events;
+	char				*name;
+};
+
+/* We can't depend on the host size for size_t, all must be 64 bit */
+typedef unsigned long long	tsize_t;
+typedef long long		stsize_t;
+
+static stsize_t
+do_write_check(struct tracecmd_output *handle, void *data, tsize_t size)
 {
 	return __do_write_check(handle->fd, data, size);
 }
 
+static short convert_endian_2(struct tracecmd_output *handle, short val)
+{
+	if (!handle->pevent)
+		return val;
+
+	return __data2host2(handle->pevent, val);
+}
+
 static int convert_endian_4(struct tracecmd_output *handle, int val)
 {
 	if (!handle->pevent)
@@ -72,6 +109,8 @@ static unsigned long long convert_endian_8(struct tracecmd_output *handle,
 
 void tracecmd_output_close(struct tracecmd_output *handle)
 {
+	int i;
+
 	if (!handle)
 		return;
 
@@ -86,6 +125,12 @@ void tracecmd_output_close(struct tracecmd_output *handle)
 	if (handle->pevent)
 		pevent_unref(handle->pevent);
 
+	if (handle->options) {
+		for (i = 0; i < handle->nr_options; i++)
+			free(handle->options[i].data);
+		free(handle->options);
+	}
+
 	free(handle);
 }
 
@@ -120,11 +165,11 @@ static unsigned long get_size(const char *file)
 	return size;
 }
 
-static unsigned long long copy_file_fd(struct tracecmd_output *handle, int fd)
+static tsize_t copy_file_fd(struct tracecmd_output *handle, int fd)
 {
-	unsigned long long size = 0;
+	tsize_t size = 0;
 	char buf[BUFSIZ];
-	int r;
+	stsize_t r;
 
 	do {
 		r = read(fd, buf, BUFSIZ);
@@ -138,10 +183,10 @@ static unsigned long long copy_file_fd(struct tracecmd_output *handle, int fd)
 	return size;
 }
 
-static unsigned long long copy_file(struct tracecmd_output *handle,
+static tsize_t copy_file(struct tracecmd_output *handle,
 				    const char *file)
 {
-	unsigned long long size = 0;
+	tsize_t size = 0;
 	int fd;
 
 	fd = open(file, O_RDONLY);
@@ -213,7 +258,7 @@ int tracecmd_ftrace_enable(int set)
 
 static int read_header_files(struct tracecmd_output *handle)
 {
-	unsigned long long size, check_size, endian8;
+	tsize_t size, check_size, endian8;
 	struct stat st;
 	char *path;
 	int fd;
@@ -292,51 +337,26 @@ static int read_header_files(struct tracecmd_output *handle)
 	return -1;
 }
 
-static int copy_event_system(struct tracecmd_output *handle, const char *sys)
+static int copy_event_system(struct tracecmd_output *handle,
+			     struct list_event_system *slist)
 {
+	struct list_event *elist;
 	unsigned long long size, check_size, endian8;
-	struct dirent *dent;
 	struct stat st;
 	char *format;
-	DIR *dir;
 	int endian4;
 	int count = 0;
 	int ret;
 
-	dir = opendir(sys);
-	if (!dir) {
-		warning("can't read directory '%s'", sys);
-		return -1;
-	}
-
-	while ((dent = readdir(dir))) {
-		if (strcmp(dent->d_name, ".") == 0 ||
-		    strcmp(dent->d_name, "..") == 0)
-			continue;
-		format = malloc_or_die(strlen(sys) + strlen(dent->d_name) + 10);
-		if (!format)
-			return -1;
-		sprintf(format, "%s/%s/format", sys, dent->d_name);
-		ret = stat(format, &st);
-		free(format);
-		if (ret < 0)
-			continue;
+	for (elist = slist->events; elist; elist = elist->next)
 		count++;
-	}
 
 	endian4 = convert_endian_4(handle, count);
 	if (do_write_check(handle, &endian4, 4))
 		return -1;
 
-	rewinddir(dir);
-	while ((dent = readdir(dir))) {
-		if (strcmp(dent->d_name, ".") == 0 ||
-		    strcmp(dent->d_name, "..") == 0)
-			continue;
-		format = malloc_or_die(strlen(sys) + strlen(dent->d_name) + 10);
-		if (!format)
-			return -1;
-		sprintf(format, "%s/%s/format", sys, dent->d_name);
+	for (elist = slist->events; elist; elist = elist->next) {
+		format = elist->file;
 		ret = stat(format, &st);
 
 		if (ret >= 0) {
@@ -344,125 +364,266 @@ static int copy_event_system(struct tracecmd_output *handle, const char *sys)
 			size = get_size(format);
 			endian8 = convert_endian_8(handle, size);
 			if (do_write_check(handle, &endian8, 8))
-				goto out_free;
+				return -1;
 			check_size = copy_file(handle, format);
 			if (size != check_size) {
 				warning("error in size of file '%s'", format);
-				goto out_free;
+				return -1;
 			}
 		}
-
-		free(format);
 	}
 
 	return 0;
+}
 
- out_free:
-	free(format);
-	return -1;
+static void add_list_event_system(struct list_event_system **systems,
+				  const char *system,
+				  const char *event,
+				  const char *path)
+{
+	struct list_event_system *slist;
+	struct list_event *elist;
+
+	for (slist = *systems; slist; slist = slist->next)
+		if (strcmp(slist->name, system) == 0)
+			break;
+
+	if (!slist) {
+		slist = malloc_or_die(sizeof(*slist));
+		slist->name = strdup(system);
+		slist->next = *systems;
+		slist->events = NULL;
+		*systems = slist;
+	}
+
+	for (elist = slist->events; elist; elist = elist->next)
+		if (strcmp(elist->name, event) == 0)
+			break;
+
+	if (!elist) {
+		elist = malloc_or_die(sizeof(*elist));
+		elist->name = strdup(event);
+		elist->file = strdup(path);
+		elist->next = slist->events;
+		slist->events = elist;
+	}
+}
+
+static void free_list_events(struct list_event_system *list)
+{
+	struct list_event_system *slist;
+	struct list_event *elist;
+
+	while (list) {
+		slist = list;
+		list = list->next;
+		while (slist->events) {
+			elist = slist->events;
+			slist->events = elist->next;
+			free(elist->name);
+			free(elist->file);
+			free(elist);
+		}
+		free(slist->name);
+		free(slist);
+	}
+}
+
+static void glob_events(struct tracecmd_output *handle,
+			struct list_event_system **systems,
+			const char *str)
+{
+	glob_t globbuf;
+	char *events_path;
+	char *system;
+	char *event;
+	char *path;
+	char *file;
+	char *ptr;
+	int do_ftrace = 0;
+	int events_len;
+	int ret;
+	int i;
+
+	if (strncmp(str, "ftrace/", 7) == 0)
+		do_ftrace = 1;
+
+	events_path = get_tracing_file(handle, "events");
+	events_len = strlen(events_path);
+
+	path = malloc_or_die(events_len + strlen(str) +
+			     strlen("/format") + 2);
+	path[0] = '\0';
+	strcat(path, events_path);
+	strcat(path, "/");
+	strcat(path, str);
+	strcat(path, "/format");
+	put_tracing_file(events_path);
+
+	globbuf.gl_offs = 0;
+	ret = glob(path, 0, NULL, &globbuf);
+	free(path);
+	if (ret < 0)
+		return;
+
+	for (i = 0; i < globbuf.gl_pathc; i++) {
+		file = globbuf.gl_pathv[i];
+		system = strdup(file + events_len + 1);
+		system = strtok_r(system, "/", &ptr);
+		if (!ptr) {
+			/* ?? should we warn? */
+			free(system);
+			continue;
+		}
+
+		if (!do_ftrace && strcmp(system, "ftrace") == 0) {
+			free(system);
+			continue;
+		}
+
+		event = strtok_r(NULL, "/", &ptr);
+		if (!ptr) {
+			/* ?? should we warn? */
+			free(system);
+			continue;
+		}
+
+		add_list_event_system(systems, system, event, file);
+		free(system);
+	}
+	globfree(&globbuf);
+}
+
+static void
+create_event_list_item(struct tracecmd_output *handle,
+		       struct list_event_system **systems,
+		       struct tracecmd_event_list *list)
+{
+	char *ptr;
+	char *str;
+
+	str = strdup(list->glob);
+	if (!str)
+		die("strdup - no memory");
+
+	/* system and event names are separated by a ':' */
+	ptr = strchr(str, ':');
+	if (ptr)
+		*ptr = '/';
+	else
+		/* system and event may also be separated by a '/' */
+		ptr = strchr(str, '/');
+
+	if (ptr) {
+		glob_events(handle, systems, str);
+		free(str);
+		return;
+	}
+
+	ptr = str;
+	str = malloc_or_die(strlen(ptr) + 3);
+	str[0] = '\0';
+	strcat(str, ptr);
+	strcat(str, "/*");
+	glob_events(handle, systems, str);
+
+	str[0] = '\0';
+	strcat(str, "*/");
+	strcat(str, ptr);
+	glob_events(handle, systems, str);
+
+	free(ptr);
+	free(str);
 }
 
 static int read_ftrace_files(struct tracecmd_output *handle)
 {
-	char *path;
+	struct list_event_system *systems = NULL;
+	struct tracecmd_event_list list = { .glob = "ftrace/*" };
 	int ret;
 
-	path = get_tracing_file(handle, "events/ftrace");
-	if (!path)
-		return -1;
+	create_event_list_item(handle, &systems, &list);
 
-	ret = copy_event_system(handle, path);
+	ret = copy_event_system(handle, systems);
 
-	put_tracing_file(path);
+	free_list_events(systems);
 
 	return ret;
 }
 
-static int read_event_files(struct tracecmd_output *handle)
+static struct list_event_system *
+create_event_list(struct tracecmd_output *handle,
+		  struct tracecmd_event_list *event_list)
 {
-	struct dirent *dent;
-	struct stat st;
-	char *path;
-	char *sys;
-	DIR *dir;
+	struct list_event_system *systems = NULL;
+	struct tracecmd_event_list *list;
+
+	for (list = event_list; list; list = list->next)
+		create_event_list_item(handle, &systems, list);
+
+	return systems;
+}
+
+static int read_event_files(struct tracecmd_output *handle,
+			    struct tracecmd_event_list *event_list)
+{
+	struct list_event_system *systems;
+	struct list_event_system *slist;
+	struct tracecmd_event_list *list;
+	struct tracecmd_event_list all_events = { .glob = "*/*" };
 	int count = 0;
 	int endian4;
 	int ret;
 
-	path = get_tracing_file(handle, "events");
-	if (!path)
-		return -1;
-
-	dir = opendir(path);
-	if (!dir)
-		die("can't read directory '%s'", path);
-
-	while ((dent = readdir(dir))) {
-		if (strcmp(dent->d_name, ".") == 0 ||
-		    strcmp(dent->d_name, "..") == 0 ||
-		    strcmp(dent->d_name, "ftrace") == 0)
-			continue;
-		ret = -1;
-		sys = malloc_or_die(strlen(path) + strlen(dent->d_name) + 2);
-		if (!sys)
-			goto out_close_dir;
-		sprintf(sys, "%s/%s", path, dent->d_name);
-		ret = stat(sys, &st);
-		free(sys);
-		if (ret < 0)
-			continue;
-		if (S_ISDIR(st.st_mode))
-			count++;
+	/*
+	 * If any of the list is the special keyword "all" then
+	 * just do all files.
+	 */
+	for (list = event_list; list; list = list->next) {
+		if (strcmp(list->glob, "all") == 0)
+			break;
 	}
+	/* all events are listed, use a global glob */
+	if (list)
+		event_list = &all_events;
+
+	systems = create_event_list(handle, event_list);
+
+	for (slist = systems; slist; slist = slist->next)
+		count++;
 
 	ret = -1;
 	endian4 = convert_endian_4(handle, count);
 	if (do_write_check(handle, &endian4, 4))
-		goto out_close_dir;
-
-	rewinddir(dir);
-	while ((dent = readdir(dir))) {
-		if (strcmp(dent->d_name, ".") == 0 ||
-		    strcmp(dent->d_name, "..") == 0 ||
-		    strcmp(dent->d_name, "ftrace") == 0)
-			continue;
-		ret = -1;
-		sys = malloc_or_die(strlen(path) + strlen(dent->d_name) + 2);
-		if (!sys)
-			goto out_close_dir;
-
-		sprintf(sys, "%s/%s", path, dent->d_name);
-		ret = stat(sys, &st);
-		if (ret >= 0) {
-			if (S_ISDIR(st.st_mode)) {
-				if (do_write_check(handle, dent->d_name,
-						   strlen(dent->d_name) + 1)) {
-					free(sys);
-					ret = -1;
-					goto out_close_dir;
-				}
-				copy_event_system(handle, sys);
-			}
-		}
-		free(sys);
-	}
-
-	put_tracing_file(path);
+		goto out_free;
 
 	ret = 0;
+	for (slist = systems; !ret && slist; slist = slist->next) {
+		if (do_write_check(handle, slist->name,
+				   strlen(slist->name) + 1)) {
+			ret = -1;
+			continue;
+		}
+		ret = copy_event_system(handle, slist);
+	}
+
+ out_free:
+	free_list_events(systems);
 
- out_close_dir:
-	closedir(dir);
 	return ret;
 }
 
-static int read_proc_kallsyms(struct tracecmd_output *handle)
+static int read_proc_kallsyms(struct tracecmd_output *handle,
+			      const char *kallsyms)
 {
 	unsigned int size, check_size, endian4;
 	const char *path = "/proc/kallsyms";
 	struct stat st;
 	int ret;
 
+	if (kallsyms)
+		path = kallsyms;
+
 	ret = stat(path, &st);
 	if (ret < 0) {
 		/* not found */
@@ -489,8 +650,8 @@ static int read_proc_kallsyms(struct tracecmd_output *handle)
 static int read_ftrace_printk(struct tracecmd_output *handle)
 {
 	unsigned int size, check_size, endian4;
-	const char *path;
 	struct stat st;
+	char *path;
 	int ret;
 
 	path = get_tracing_file(handle, "printk_formats");
@@ -503,25 +664,33 @@ static int read_ftrace_printk(struct tracecmd_output *handle)
 		size = 0;
 		endian4 = convert_endian_4(handle, size);
 		if (do_write_check(handle, &endian4, 4))
-			return -1;
-		return 0;
+			goto fail;
+		goto out;
 	}
 	size = get_size(path);
 	endian4 = convert_endian_4(handle, size);
 	if (do_write_check(handle, &endian4, 4))
-		return -1;
+		goto fail;
 	check_size = copy_file(handle, path);
 	if (size != check_size) {
 		errno = EINVAL;
 		warning("error in size of file '%s'", path);
-		return -1;
+		goto fail;
 	}
 
+ out:
+	put_tracing_file(path);
 	return 0;
+ fail:
+	put_tracing_file(path);
+	return -1;
 }
 
 static struct tracecmd_output *
-create_file_fd(int fd, int cpus, struct tracecmd_input *ihandle)
+create_file_fd(int fd, struct tracecmd_input *ihandle,
+	       const char *tracing_dir,
+	       const char *kallsyms,
+	       struct tracecmd_event_list *list)
 {
 	struct tracecmd_output *handle;
 	unsigned long long endian8;
@@ -540,6 +709,11 @@ create_file_fd(int fd, int cpus, struct tracecmd_input *ihandle)
 	memset(handle, 0, sizeof(*handle));
 
 	handle->fd = fd;
+	if (tracing_dir) {
+		handle->tracing_dir = strdup(tracing_dir);
+		if (!handle->tracing_dir)
+			goto out_free;
+	}
 
 	buf[0] = 23;
 	buf[1] = 8;
@@ -590,9 +764,9 @@ create_file_fd(int fd, int cpus, struct tracecmd_input *ihandle)
 		goto out_free;
 	if (read_ftrace_files(handle))
 		goto out_free;
-	if (read_event_files(handle))
+	if (read_event_files(handle, list))
 		goto out_free;
-	if (read_proc_kallsyms(handle))
+	if (read_proc_kallsyms(handle, kallsyms))
 		goto out_free;
 	if (read_ftrace_printk(handle))
 		goto out_free;
@@ -629,8 +803,11 @@ create_file_fd(int fd, int cpus, struct tracecmd_input *ihandle)
 	return NULL;
 }
 
-static struct tracecmd_output *create_file(const char *output_file, int cpus,
-					   struct tracecmd_input *ihandle)
+static struct tracecmd_output *create_file(const char *output_file,
+					   struct tracecmd_input *ihandle,
+					   const char *tracing_dir,
+					   const char *kallsyms,
+					   struct tracecmd_event_list *list)
 {
 	struct tracecmd_output *handle;
 	int fd;
@@ -639,7 +816,7 @@ static struct tracecmd_output *create_file(const char *output_file, int cpus,
 	if (fd < 0)
 		return NULL;
 
-	handle = create_file_fd(fd, cpus, ihandle);
+	handle = create_file_fd(fd, ihandle, tracing_dir, kallsyms, list);
 	if (!handle) {
 		close(fd);
 		unlink(output_file);
@@ -648,23 +825,79 @@ static struct tracecmd_output *create_file(const char *output_file, int cpus,
 	return handle;
 }
 
+/**
+ * tracecmd_add_option - add options to the file
+ * @handle: the output file handle name
+ * @id: the id of the option
+ * @size: the size of the option data
+ * @data: the data to write to the file.
+ */
+int tracecmd_add_option(struct tracecmd_output *handle,
+			unsigned short id,
+			int size, void *data)
+{
+	int index = handle->nr_options;
+
+	/*
+	 * We can only add options before they were written.
+	 * This may change in the future.
+	 */
+	if (handle->options_written)
+		return -EBUSY;
+
+	handle->nr_options++;
+
+	if (!handle->options)
+		handle->options = malloc_or_die(sizeof(*handle->options));
+	else {
+		handle->options = realloc(handle->options,
+					  sizeof(*handle->options) * handle->nr_options);
+		if (!handle->options)
+			die("Could not reallocate space for options");
+	}
+
+	handle->options[index].id = id;
+	handle->options[index].size = size;
+	handle->options[index].data = malloc_or_die(size);
+	memcpy(handle->options[index].data, data, size);
+
+	return 0;
+}
+
 static int add_options(struct tracecmd_output *handle)
 {
 	unsigned short option;
+	unsigned short endian2;
+	unsigned int endian4;
+	int i;
+
+	if (handle->options_written)
+		die("options already written?");
 
 	if (do_write_check(handle, "options  ", 10))
 		return -1;
 
-	/*
-	 * Right now we have no options, but this is where options
-	 * will be added in the future.
-	 */
+	for (i = 0; i < handle->nr_options; i++) {
+		endian2 = convert_endian_2(handle, handle->options[i].id);
+		if (do_write_check(handle, &endian2, 2))
+			return -1;
+
+		endian4 = convert_endian_4(handle, handle->options[i].size);
+		if (do_write_check(handle, &endian4, 4))
+			return -1;
+
+		if (do_write_check(handle, handle->options[i].data,
+				   handle->options[i].size))
+			return -1;
+	}
 
 	option = TRACECMD_OPTION_DONE;
 
 	if (do_write_check(handle, &option, 2))
 		return -1;
 
+	handle->options_written = 1;
+
 	return 0;
 }
 
@@ -673,7 +906,7 @@ struct tracecmd_output *tracecmd_create_file_latency(const char *output_file, in
 	struct tracecmd_output *handle;
 	char *path;
 
-	handle = create_file(output_file, cpus, NULL);
+	handle = create_file(output_file, NULL, NULL, NULL, &all_event_list);
 	if (!handle)
 		return NULL;
 
@@ -705,9 +938,9 @@ out_free:
 int tracecmd_append_cpu_data(struct tracecmd_output *handle,
 			     int cpus, char * const *cpu_data_files)
 {
-	unsigned long long *offsets = NULL;
+	off64_t *offsets = NULL;
 	unsigned long long *sizes = NULL;
-	unsigned long long offset;
+	off64_t offset;
 	unsigned long long endian8;
 	off64_t check_size;
 	char *file;
@@ -733,7 +966,7 @@ int tracecmd_append_cpu_data(struct tracecmd_output *handle,
 	if (!sizes)
 		goto out_free;
 
-	offset = lseek(handle->fd, 0, SEEK_CUR);
+	offset = lseek64(handle->fd, 0, SEEK_CUR);
 
 	/* hold any extra data for data */
 	offset += cpus * (16);
@@ -760,9 +993,10 @@ int tracecmd_append_cpu_data(struct tracecmd_output *handle,
 	}
 
 	for (i = 0; i < cpus; i++) {
-		fprintf(stderr, "offset=%llx\n", offsets[i]);
-		ret = lseek64(handle->fd, offsets[i], SEEK_SET);
-		if (ret < 0) {
+		fprintf(stderr, "CPU%d data recorded at offset=0x%llx\n",
+			i, (unsigned long long) offsets[i]);
+		offset = lseek64(handle->fd, offsets[i], SEEK_SET);
+		if (offset == (off64_t)-1) {
 			warning("could not seek to %lld\n", offsets[i]);
 			goto out_free;
 		}
@@ -773,6 +1007,7 @@ int tracecmd_append_cpu_data(struct tracecmd_output *handle,
 			    check_size, sizes[i]);
 			goto out_free;
 		}
+		fprintf(stderr, "    %llu bytes in size\n", (unsigned long long)check_size);
 	}
 
 	free(offsets);
@@ -783,8 +1018,6 @@ int tracecmd_append_cpu_data(struct tracecmd_output *handle,
  out_free:
 	free(offsets);
 	free(sizes);
-
-	tracecmd_output_close(handle);
 	return -1;
 }
 
@@ -812,7 +1045,7 @@ int tracecmd_attach_cpu_data_fd(int fd, int cpus, char * const *cpu_data_files)
 
 	handle = malloc(sizeof(*handle));
 	if (!handle)
-		return -1;
+		goto out_free;
 	memset(handle, 0, sizeof(*handle));
 
 	handle->fd = fd;
@@ -824,10 +1057,9 @@ int tracecmd_attach_cpu_data_fd(int fd, int cpus, char * const *cpu_data_files)
 	pevent_ref(pevent);
 	handle->page_size = tracecmd_page_size(ihandle);
 
-	if (tracecmd_append_cpu_data(handle, cpus, cpu_data_files) < 0)
-		goto out_free;
+	if (tracecmd_append_cpu_data(handle, cpus, cpu_data_files) >= 0)
+		ret = 0;
 
-	ret = 0;
 	tracecmd_output_close(handle);
  out_free:
 	tracecmd_close(ihandle);
@@ -845,31 +1077,60 @@ int tracecmd_attach_cpu_data(char *file, int cpus, char * const *cpu_data_files)
 	return tracecmd_attach_cpu_data_fd(fd, cpus, cpu_data_files);
 }
 
-struct tracecmd_output *tracecmd_create_file(const char *output_file,
-					     int cpus, char * const *cpu_data_files)
+struct tracecmd_output *
+tracecmd_create_file_glob(const char *output_file,
+			  int cpus, char * const *cpu_data_files,
+			  struct tracecmd_event_list *list)
 {
 	struct tracecmd_output *handle;
 
-	handle = create_file(output_file, cpus, NULL);
+	handle = create_file(output_file, NULL, NULL, NULL, list);
 	if (!handle)
 		return NULL;
 
-	if (tracecmd_append_cpu_data(handle, cpus, cpu_data_files) < 0)
+	if (tracecmd_append_cpu_data(handle, cpus, cpu_data_files) < 0) {
+		tracecmd_output_close(handle);
 		return NULL;
+	}
 
 	return handle;
 }
 
-struct tracecmd_output *
-tracecmd_create_init_fd(int fd, int cpus)
+struct tracecmd_output *tracecmd_create_file(const char *output_file,
+					     int cpus, char * const *cpu_data_files)
 {
-	struct tracecmd_output *handle;
+	return tracecmd_create_file_glob(output_file, cpus,
+					 cpu_data_files, &all_event_list);
+}
 
-	handle = create_file_fd(fd, cpus, NULL);
-	if (!handle)
-		return NULL;
+struct tracecmd_output *tracecmd_create_init_fd(int fd)
+{
+	return create_file_fd(fd, NULL, NULL, NULL, &all_event_list);
+}
 
-	return handle;
+struct tracecmd_output *
+tracecmd_create_init_fd_glob(int fd, struct tracecmd_event_list *list)
+{
+	return create_file_fd(fd, NULL, NULL, NULL, list);
+}
+
+struct tracecmd_output *
+tracecmd_create_init_file_glob(const char *output_file,
+			       struct tracecmd_event_list *list)
+{
+	return create_file(output_file, NULL, NULL, NULL, list);
+}
+
+struct tracecmd_output *tracecmd_create_init_file(const char *output_file)
+{
+	return create_file(output_file, NULL, NULL, NULL, &all_event_list);
+}
+
+struct tracecmd_output *tracecmd_create_init_file_override(const char *output_file,
+							   const char *tracing_dir,
+							   const char *kallsyms)
+{
+	return create_file(output_file, NULL, tracing_dir, kallsyms, &all_event_list);
 }
 
 /**
@@ -886,7 +1147,7 @@ struct tracecmd_output *tracecmd_copy(struct tracecmd_input *ihandle,
 {
 	struct tracecmd_output *handle;
 
-	handle = create_file(file, tracecmd_cpus(ihandle), ihandle);
+	handle = create_file(file, ihandle, NULL, NULL, &all_event_list);
 	if (!handle)
 		return NULL;
 
